import sys
import os
import asyncio
import uuid
import pytest
import json
import warnings
from pathlib import Path
from typing import Dict, Any, List, Optional
from pydantic import BaseModel, Field
from tqdm.asyncio import tqdm_asyncio

# Suppress multiprocessing resource tracker warnings
warnings.filterwarnings("ignore", message=".*leaked semaphore.*", category=UserWarning)

from examples.verilog.environment import VerilogEnvironment
from examples.verilog.taskset import VerilogTaskInstance, create_verilog_taskset
from examples.verilog.engine import VerilogPrivateState, VerilogPublicState
from environment.tools import EnvToolCall
from tasks.core import Impetus, Intent
from synth_ai.zyk import LM


# Tool argument models for the agent
class WriteFileArgs(BaseModel):
    path: str = Field(description="Path to the Verilog file to write")
    content: str = Field(description="Verilog code content")
    reasoning: str = Field(description="Reasoning for the code implementation")


class CompileArgs(BaseModel):
    sources: Optional[List[str]] = Field(None, description="List of source files to compile")
    testbench: Optional[str] = Field(None, description="Testbench file to include")
    reasoning: str = Field(description="Reasoning for compilation step")


class SimulateArgs(BaseModel):
    binary: Optional[str] = Field(None, description="Binary file to simulate")
    reasoning: str = Field(description="Reasoning for simulation step")


class RunCustomTestArgs(BaseModel):
    test_file: str = Field(description="Path to the custom test file")
    design_files: Optional[List[str]] = Field(None, description="List of design files to test against")
    reasoning: str = Field(description="Reasoning for running custom test")


class SubmitArgs(BaseModel):
    reasoning: str = Field(description="Reasoning for submission")


class TerminateArgs(BaseModel):
    reason: str = Field(description="Reason for termination")


# Environment tool call wrappers
class WriteFile(EnvToolCall):
    def __init__(self, path: str, content: str):
        super().__init__(tool="write_file", args={"path": path, "content": content})


class Compile(EnvToolCall):
    def __init__(self, sources: Optional[List[str]] = None, testbench: Optional[str] = None):
        super().__init__(tool="compile", args={"sources": sources, "testbench": testbench})


class Simulate(EnvToolCall):
    def __init__(self, binary: Optional[str] = None):
        super().__init__(tool="simulate", args={"binary": binary})


class RunCustomTest(EnvToolCall):
    def __init__(self, test_file: str, design_files: Optional[List[str]] = None):
        super().__init__(tool="run_custom_test", args={"test_file": test_file, "design_files": design_files})


class Submit(EnvToolCall):
    def __init__(self):
        super().__init__(tool="submit", args={})


def format_obs_for_llm(obs: Dict[str, Any]) -> str:
    """Format observation for LLM input."""
    files_info = ""
    if obs.get("files"):
        files_info = "Available files:\n"
        for filename, content in obs["files"].items():
            files_info += f"  {filename}:\n"
            # Show first few lines of content
            lines = content.split('\n')[:10]
            for line in lines:
                files_info += f"    {line}\n"
            if len(content.split('\n')) > 10:
                files_info += "    ...\n"
        files_info += "\n"
    
    compile_status = obs.get("compile_status", "")
    simulate_status = obs.get("simulate_status", "")
    
    status_info = f"Task completed: {obs.get('task_completed', False)}\n"
    status_info += f"Terminated: {obs.get('terminated', False)}\n"
    status_info += f"Total reward: {obs.get('total_reward', 0)}\n"
    status_info += f"Last reward: {obs.get('reward_last', 0)}\n"
    
    
    if compile_status:
        status_info += f"Compile status: {compile_status}\n"
    else:
        status_info += "Compile status: No compilation output\n"
        
    if simulate_status:
        status_info += f"Simulate status: {simulate_status}\n"
    else:
        status_info += "Simulate status: No simulation output\n"
    
    custom_test_status = obs.get("custom_test_status", "")
    if custom_test_status:
        status_info += f"Custom test status: {custom_test_status}\n"
    else:
        status_info += "Custom test status: No custom tests run\n"
    
    return f"{files_info}{status_info}"


class VerilogReActAgent:
    """Simple ReAct agent for Verilog tasks."""
    
    def __init__(self, llm, max_turns: int = 10):
        self.llm = llm
        self.max_turns = max_turns
        self.history: List[Dict[str, Any]] = []
        self.task_description = ""
        self.system_name = "verilog-react"
        self.system_instance_id = str(uuid.uuid4())
        
        self.tools = [
            {
                "type": "function",
                "function": {
                    "name": "write_file",
                    "description": "Write Verilog code to a file",
                    "parameters": WriteFileArgs.model_json_schema(),
                },
            },
            {
                "type": "function", 
                "function": {
                    "name": "compile",
                    "description": "Compile Verilog sources with iverilog",
                    "parameters": CompileArgs.model_json_schema(),
                },
            },
            {
                "type": "function",
                "function": {
                    "name": "simulate", 
                    "description": "Run simulation with vvp",
                    "parameters": SimulateArgs.model_json_schema(),
                },
            },
            {
                "type": "function",
                "function": {
                    "name": "run_custom_test",
                    "description": "Compile and run agent-written test against design files",
                    "parameters": RunCustomTestArgs.model_json_schema(),
                },
            },
            {
                "type": "function",
                "function": {
                    "name": "submit",
                    "description": "Submit solution for grading",
                    "parameters": SubmitArgs.model_json_schema(),
                },
            },
            {
                "type": "function",
                "function": {
                    "name": "terminate",
                    "description": "Terminate if task is complete or cannot proceed",
                    "parameters": TerminateArgs.model_json_schema(),
                },
            },
        ]

    def set_task_description(self, description: str):
        """Set the task description for this agent."""
        self.task_description = description

    async def decide(self, obs: str) -> Dict[str, Any]:
        """Decide next action based on observation."""
        self.history.append({"type": "observation", "content": obs})
        
        # Build prompt from history
        history_text = ""
        for entry in self.history[-5:]:  # Last 5 entries
            if entry["type"] == "observation":
                history_text += f"OBSERVATION:\n{entry['content']}\n\n"
            elif entry["type"] == "tool_call":
                history_text += f"ACTION: Called {entry['tool_name']} with args: {entry['tool_args']}\n\n"
            elif entry["type"] == "tool_response":
                history_text += f"RESULT: {entry['content']}\n\n"
        
        prompt = f"""Task: {self.task_description}

History:
{history_text}

Based on the observation and history, decide what to do next.

Note - compiling theerilog

Choose the most appropriate tool to call next."""

        system_message = """You are a Verilog design expert. Your goal is to implement correct Verilog code that passes testbenches.

Available tools:
- write_file: Write Verilog code to files
- compile: Compile Verilog sources with iverilog  
- simulate: Run simulation with vvp
- run_custom_test: Write and run your own tests against design files
- submit: Submit solution when complete
- terminate: End if task complete or cannot proceed

Note: You can write custom testbenches to verify your design logic before running the official testbench.
Always use the tools available. Include reasoning in your tool calls."""

        try:
            response = await self.llm.respond_async(
                system_message=system_message,
                user_message=prompt,
                tools=self.tools
            )
            
            if not response.tool_calls:
                return {"action": "terminate", "args": {"reason": "No tool call generated"}}
                
            tool_call = response.tool_calls[0]
            
            # Handle different response structures
            if hasattr(tool_call, 'function'):
                # Standard OpenAI format
                tool_name = tool_call.function.name
                tool_args = tool_call.function.arguments
            elif isinstance(tool_call, dict):
                # Dictionary format
                if 'function' in tool_call:
                    tool_name = tool_call['function']['name']
                    tool_args = tool_call['function']['arguments']
                else:
                    tool_name = tool_call.get('name', 'unknown')
                    tool_args = tool_call.get('arguments', {})
            else:
                return {"action": "terminate", "args": {"reason": f"Unexpected tool call format: {type(tool_call)}"}}
            
            if isinstance(tool_args, str):
                import json
                tool_args = json.loads(tool_args)
                
            self.history.append({
                "type": "tool_call", 
                "tool_name": tool_name,
                "tool_args": tool_args
            })
            
            return {"action": tool_name, "args": tool_args}
            
        except Exception as e:
            import traceback
            traceback.print_exc()
            return {"action": "terminate", "args": {"reason": f"Error: {str(e)}"}}


async def run_verilog_episode(task_instance: VerilogTaskInstance, model_name: str, debug: bool = False) -> bool:
    """Run a single episode with the Verilog environment and agent."""
    
    task_name = task_instance.metadata.problem_name
    if debug:
        print(f"[DEBUG] Starting episode for task: {task_name}")
    
    # Create environment
    env = VerilogEnvironment(task_instance)
    
    # Create agent
    llm = LM(model_name=model_name, formatting_model_name=model_name, temperature=0.0)
    agent = VerilogReActAgent(llm)
    
    # Set task description from the task instance
    agent.set_task_description(task_instance.impetus.instructions)
    if debug:
        print(f"[DEBUG] Task description: {task_instance.impetus.instructions}")
    
    try:
        # Initialize environment
        if debug:
            print(f"[DEBUG] Initializing environment...")
        obs = await env.initialize()
        obs_text = format_obs_for_llm(obs)
        if debug:
            print(f"[DEBUG] Initial observation: {obs_text[:200]}...")
        
        # Run episode
        for turn in range(agent.max_turns):
            if debug:
                print(f"[DEBUG] Turn {turn + 1}/{agent.max_turns}")
            
            # Agent decides action
            decision = await agent.decide(obs_text)
            if debug:
                print(f"[DEBUG] Agent decision: {decision}")
            
            if decision["action"] == "terminate":
                reason = decision["args"].get("reason", "Agent terminated")
                agent.history.append({"type": "tool_response", "content": f"Terminated: {reason}"})
                if debug:
                    print(f"[DEBUG] Agent terminated: {reason}")
                break
                
            # Execute action in environment
            action_name = decision["action"]
            action_args = decision["args"]
            
            # Create appropriate tool call
            if action_name == "write_file":
                tool_call = WriteFile(action_args["path"], action_args["content"])
            elif action_name == "compile":
                tool_call = Compile(action_args.get("sources"), action_args.get("testbench"))
            elif action_name == "simulate":
                tool_call = Simulate(action_args.get("binary"))
            elif action_name == "run_custom_test":
                tool_call = RunCustomTest(action_args["test_file"], action_args.get("design_files"))
            elif action_name == "submit":
                tool_call = Submit()
            else:
                agent.history.append({"type": "tool_response", "content": f"Unknown action: {action_name}"})
                if debug:
                    print(f"[DEBUG] Unknown action: {action_name}")
                continue
                
            # Step environment
            if debug:
                print(f"[DEBUG] Stepping environment with {action_name}")
            obs = await env.step(tool_call)
            obs_text = format_obs_for_llm(obs)
            if debug:
                print(f"[DEBUG] Environment response: {obs_text[:200]}...")
            
            # Record result
            agent.history.append({"type": "tool_response", "content": obs_text})
            
            # Check if terminated
            if obs.get("terminated", False):
                task_completed = obs.get("task_completed", False)
                if debug:
                    print(f"[DEBUG] Environment terminated. Task completed: {task_completed}")
                    print(f"[DEBUG] Final observation: {obs}")
                return task_completed
                
        if debug:
            print(f"[DEBUG] Episode ended after {agent.max_turns} turns without completion")
            print(f"[DEBUG] Final observation: {obs}")
            print(f"[DEBUG] Agent history length: {len(agent.history)}")
        return False
        
    except Exception as e:
        print(f"[ERROR] Episode failed with error: {e}")
        import traceback
        traceback.print_exc()
        return False


async def eval_verilog_react(
    model_name: str = "gpt-4.1-nano",
    formatting_model_name: str = "gpt-4.1-nano",
    n_instances: int = 1,
    multiplicity: int = 1,
    debug_mode=False
) -> Dict[str, Any]:
    """Evaluate the ReAct agent on Verilog tasks."""
    
    # Create task set
    taskset = await create_verilog_taskset(max_instances=n_instances)
    
    print(f"SYNTH: Starting Verilog ReAct Agent Evaluation for Model: {model_name}")
    print(f"Running {n_instances} instances...")
    
    # Run multiple instances of each task
    all_results = []
    
    # Use tqdm to track task completion, not instance completion
    from tqdm import tqdm
    
    for task_instance in tqdm(taskset.instances, desc="Tasks", unit="task"):
        task_name = task_instance.metadata.problem_name
        
        # Create coroutines for all instances of this task
        task_coroutines = [
            run_verilog_episode(task_instance, model_name, debug=debug_mode)
            for _ in range(multiplicity)
        ]
        
        # Run all instances in parallel without progress tracking
        task_results = await asyncio.gather(*task_coroutines)
        
        # Calculate success rate for this task
        success_count = sum(task_results)
        success_rate = success_count / len(task_results)
        
        all_results.append({
            "task": task_name,
            "difficulty": task_instance.metadata.difficulty,
            "success_count": success_count,
            "total_instances": len(task_results),
            "success_rate": success_rate
        })
    
    # Calculate overall statistics
    total_successes = sum(r["success_count"] for r in all_results)
    total_attempts = sum(r["total_instances"] for r in all_results)
    overall_success_rate = total_successes / total_attempts if total_attempts > 0 else 0.0
    
    return {
        "model": model_name,
        "total_successes": total_successes,
        "total_attempts": total_attempts,
        "overall_success_rate": overall_success_rate,
        "task_results": all_results
    }


@pytest.mark.asyncio
async def test_verilog_react_agent():
    """Test the Verilog ReAct agent on a simple task."""
    
    # Create a simple task set
    taskset = await create_verilog_taskset()
    
    # Test with the first task (should be the adder)
    task_instance = taskset.instances[0]
    
    # Run episode
    success = await run_verilog_episode(task_instance, "gpt-4.1-nano")
    
    print(f"Task: {task_instance.metadata.problem_name}")
    print(f"Success: {success}")
    
    # For testing, we'll allow failure since this is a basic implementation
    assert success or not success  # Always pass for now


# async def debug_single_run():
#     """Debug a single run to understand what's happening."""
#     from tabulate import tabulate
    
#     print("Starting debug run with gpt-4.1...")
    
#     # Run single evaluation with debugging
#     result = await eval_verilog_react(
#         model_name="gpt-4.1",
#         formatting_model_name="gpt-4.1",
#         n_instances=1  # Just 1 instance for debugging
#     )
    
#     print("\n=== DEBUG EVALUATION COMPLETED ===")
#     print(f"Model: {result['model']}")
#     print(f"Total Successes: {result['total_successes']}")
#     print(f"Total Attempts: {result['total_attempts']}")
#     print(f"Success Rate: {result['overall_success_rate']:.1%}")
    
#     for task_result in result["task_results"]:
#         print(f"\nTask: {task_result['task']}")
#         print(f"  Difficulty: {task_result['difficulty']}")
#         print(f"  Success Rate: {task_result['success_rate']:.1%}")
#         print(f"  Successes: {task_result['success_count']}/{task_result['total_instances']}")


async def run_parallel_evaluation(
        models_to_test = ["gpt-4.1-nano", "gpt-4.1-mini"],
        n_instances=3,
        multiplicity=1
):
    """Run evaluation for all models in parallel."""
    from tabulate import tabulate

    print(f"Starting parallel evaluation for {len(models_to_test)} models...")
    print(f"Models: {', '.join(models_to_test)}")
    
    # Create coroutines for all model evaluations
    model_coroutines = [
        eval_verilog_react(
            model_name=model_name,
            formatting_model_name=model_name,
            n_instances=n_instances,
            multiplicity=multiplicity
        )
        for model_name in models_to_test
    ]
    
    # Run all model evaluations in parallel
    results_from_all_models = await asyncio.gather(*model_coroutines)
    
    print("\n=== PARALLEL EVALUATION COMPLETED ===")
    
    # Create summary table
    summary_data = []
    for result in results_from_all_models:
        summary_data.append({
            "Model": result["model"],
            "Total Successes": result["total_successes"],
            "Total Attempts": result["total_attempts"],
            "Overall Success Rate": f"{result['overall_success_rate']:.1%}"
        })
    
    print("\n--- Model Comparison Summary ---")
    print(tabulate(summary_data, headers="keys", tablefmt="github"))
    
    # Detailed breakdown by task
    print("\n--- Detailed Results by Task ---")
    for result in results_from_all_models:
        print(f"\n{result['model']}:")
        task_data = []
        for task_result in result["task_results"]:
            task_data.append({
                "Task": task_result["task"],
                "Difficulty": task_result["difficulty"],
                "Success Rate": f"{task_result['success_rate']:.1%}",
                "Successes": f"{task_result['success_count']}/{task_result['total_instances']}"
            })
        print(tabulate(task_data, headers="keys", tablefmt="github"))


if __name__ == "__main__":
    import argparse
    
    asyncio.run(
        run_parallel_evaluation(
            models_to_test=["gpt-4.1-nano"],  # , "gpt-4.1-mini", "gpt-4.1"
            n_instances=30,
            multiplicity=1
        )
    )
    # mini: 26/30 synth
    # nano: 0/0 synth